@@ -28,29 +28,31 @@ import datetime
 from distutils import dir_util
 import re
 import sqlite3
-try:
-    import thread
-except ImportError:
-    thread = None
+import json
+import uuid
 
 from datatypes import Datetime
 
 # first party
-from ..exception import UniqueError
+from ..exception import (
+    InterfaceError,
+    UniqueError,
+    TableError,
+    FieldError,
+    UniqueError,
+    CloseError,
+)
+
 from ..compat import *
 from .base import SQLInterface, SQLConnection
 
 
 class SQLiteRowDict(sqlite3.Row):
-    def __getitem__(self, k):
-        if is_py2:
-            return super(SQLiteRowDict, self).__getitem__(b"{}".format(k))
-        else:
-            return super(SQLiteRowDict, self).__getitem__(k)
-
     def get(self, k, default_val=None):
-        r = default_val
-        r = self[k]
+        try:
+            r = self[k]
+        except KeyError:
+            r = default_val
         return r
 
 
@@ -61,16 +63,39 @@ class SQLiteConnection(SQLConnection, sqlite3.Connection):
     parent class
     """
     def __init__(self, *args, **kwargs):
-        super(SQLiteConnection, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.closed = 0
 
     def close(self, *args, **kwargs):
-        r = super(SQLiteConnection, self).close(*args, **kwargs)
+        r = super().close(*args, **kwargs)
         self.closed = 1
         return r
 
 
-class TimestampType(object):
+class BooleanType(object):
+    FIELD_TYPE = 'BOOL'
+
+    @classmethod
+    def adapt(cls, val):
+        """From python you get False and True, convert those to 1/0"""
+        return 1 if val else 0
+
+    @classmethod
+    def convert(cls, val):
+        """from the db you get values like b'0' and b'1', convert those to True/False"""
+        return bool(int(val))
+
+
+class StringType(object):
+    """this just makes sure 8-bit bytestrings get converted ok"""
+    FIELD_TYPE = 'TEXT'
+
+    @classmethod
+    def adapt(cls, val):
+        return String(val)
+
+
+class DatetimeType(StringType):
     """External sqlite3 databases can store the TIMESTAMP type as unix timestamps,
     this caused parsing problems when pulling the values out of the db because the
     default adapter expected TIMESTAMP to be in the form of YYYY-MM-DD HH:MM:SS.SSSSSS
@@ -79,113 +104,78 @@ class TimestampType(object):
     https://www.sqlite.org/lang_datefunc.html
     the "unixepoch" modifier only works for dates between 0000-01-01 00:00:00 and
     5352-11-01 10:52:47 (unix times of -62167219200 through 106751991167)
+
+    uses the name TIMESTAMP over DATETIME to be consistent with Postgres
     """
-    @staticmethod
-    def adapt(val):
-        return val.isoformat(b" ") if is_py2 else val.isoformat(" ")
+    FIELD_TYPE = 'TIMESTAMP'
 
-    @staticmethod
-    def convert(val):
-        val = StringType.adapt(val)
-        if re.match(r"^\-?\d+\.\d+$", val):
-            # account for unix timestamps with microseconds
-            val = datetime.datetime.fromtimestamp(float(val))
+    @classmethod
+    def adapt(cls, val):
+        return Datetime(val).isoformat()
 
-        elif re.match(r"^\-?\d+$", val):
-            # account for unix timestamps without microseconds
-            val = int(val)
-
-            try:
-                val = datetime.datetime.fromtimestamp(val)
-
-            except ValueError:
-                # we're hosed with this unix timestamp, but rather than error
-                # out let's go ahead and return the closest approximation we
-                # can get to the correct timestamp
-                if val > 0:
-                    val = datetime.datetime.max
-                else:
-                    val = datetime.datetime.min
-
-        else:
-            # ISO 8601 is not very strict with the date format and this tries to
-            # capture most of that leniency, with the one exception that the
-            # date must be in UTC
-            # https://en.wikipedia.org/wiki/ISO_8601
-            m = re.match(
-                r"^(\d{4}).?(\d{2}).?(\d{2}).(\d{2}):?(\d{2}):?(\d{2})(?:\.(\d+))?Z?$",
-                val
-            )
-
-            if m:
-                parsed_dateparts = m.groups()
-                dateparts = list(map(lambda x: int(x) if x else 0, parsed_dateparts[:6]))
-                val = datetime.datetime(*dateparts)
-
-                # account for ms with leading zeros
-                if parsed_dateparts[6]:
-                    ms_len = len(parsed_dateparts[6])
-                    if ms_len >= 3:
-                        millis = parsed_dateparts[6][:3]
-                        micros = parsed_dateparts[6][3:] or 0
-
-                    else:
-                        millis = parsed_dateparts[6] or 0
-                        micros = 0
-
-                    # make sure each part is 3 digits by zero padding on the right
-                    if millis:
-                        millis = "{:0<3.3}".format(millis)
-                    if micros:
-                        micros = "{:0<3.3}".format(micros)
-
-                    val += datetime.timedelta(
-                        milliseconds=int(millis),
-                        microseconds=int(micros)
-                    )
-
-            else:
-                raise ValueError("Cannot infer UTC datetime value from {}".format(val))
-
-        return val
-
-
-class BooleanType(object):
-    @staticmethod
-    def adapt(val):
-        """From python you get False and True, convert those to 1/0"""
-        return 1 if val else 0
-
-    @staticmethod
-    def convert(val):
-        """from the db you get values like b'0' and b'1', convert those to True/False"""
-        return bool(int(val))
+    @classmethod
+    def convert(cls, val):
+        return Datetime(super().adapt(val)).datetime()
+        #return Datetime(super().adapt(val)).datetime().replace(tzinfo=datetime.timezone.utc)
 
 
 class NumericType(object):
+    """numbers bigger than 64bit integers can be stored as this
+
+    This is named to be as consistent with Postgres's NUMERIC(<precision>, 0) type
+
+    This has TEXT in the name so it is treated as text according to SQLite's
+    order of affinity rule 2:
+
+        If the declared type of the column contains any of the strings "CHAR",
+        "CLOB", or "TEXT" then that column has TEXT affinity.
+
+        https://www.sqlite.org/datatype3.html
+    """
+    FIELD_TYPE = 'NUMERICTEXT'
+
+    @classmethod
+    def adapt(cls, val):
+        if val < 9223372036854775807:
+            return val
+        else:
+            v = str(val)
+            return v
+
+    @classmethod
+    def convert(cls, val):
+        """This should only be called when the column type is actually FIELD_TYPE"""
+        return int(val)
+
+
+class DecimalType(object):
+    FIELD_TYPE = 'DECIMALTEXT'
+
     @staticmethod
     def adapt(val):
-        return float(str(val))
+        return str(val)
 
     @staticmethod
     def convert(val):
-        if is_py2:
-            ret = decimal.Decimal(str(val))
-        else:
-            val = StringType.adapt(val)
-            ret = decimal.Decimal(val)
+        ret = decimal.Decimal(val)
         return ret
 
 
-class StringType(object):
-    """this just makes sure 8-bit bytestrings get converted ok"""
-    @staticmethod
-    def adapt(val):
-        #if isinstance(val, str):
-        if isinstance(val, bytes):
-            val = val.decode('utf-8')
+class DictType(object):
+    """Converts a dict to json text and back again
 
-        return val
+    Uses JSONBTEXT to be as close to Postgres while still triggering SQLite's affinity rule 2
+    """
+    FIELD_TYPE = 'JSONBTEXT'
+
+    @classmethod
+    def adapt(cls, val):
+        return json.dumps(val)
+
+    @classmethod
+    def convert(cls, val):
+        """This should only be called when the column type is actually FIELD_TYPE"""
+        return json.loads(val)
 
 
 class SQLite(SQLInterface):
@@ -253,19 +243,28 @@ class SQLite(SQLInterface):
 
         # for some reason this is needed in python 3.6 in order for saved bytes
         # to be ran through the converter, not sure why
-        sqlite3.register_converter(b'TEXT' if is_py2 else 'TEXT', StringType.adapt)
+        sqlite3.register_converter(StringType.FIELD_TYPE, StringType.adapt)
 
-        sqlite3.register_adapter(decimal.Decimal, NumericType.adapt)
-        sqlite3.register_converter(b'NUMERIC' if is_py2 else 'NUMERIC', NumericType.convert)
+        #sqlite3.register_adapter(decimal.Decimal, NumericType.adapt)
+        #sqlite3.register_converter('NUMERIC', NumericType.convert)
 
         sqlite3.register_adapter(bool, BooleanType.adapt)
-        sqlite3.register_converter(b'BOOLEAN' if is_py2 else 'BOOLEAN', BooleanType.convert)
+        sqlite3.register_converter(BooleanType.FIELD_TYPE, BooleanType.convert)
 
         # sadly, it doesn't look like these work for child classes so each class
         # has to be adapted even if its parent is already registered
-        sqlite3.register_adapter(datetime.datetime, TimestampType.adapt)
-        sqlite3.register_adapter(Datetime, TimestampType.adapt)
-        sqlite3.register_converter(b'TIMESTAMP' if is_py2 else 'TIMESTAMP', TimestampType.convert)
+        sqlite3.register_adapter(datetime.datetime, DatetimeType.adapt)
+        sqlite3.register_adapter(Datetime, DatetimeType.adapt)
+        sqlite3.register_converter(DatetimeType.FIELD_TYPE, DatetimeType.convert)
+
+        sqlite3.register_adapter(int, NumericType.adapt)
+        sqlite3.register_converter(NumericType.FIELD_TYPE, NumericType.convert)
+
+        sqlite3.register_adapter(decimal.Decimal, DecimalType.adapt)
+        sqlite3.register_converter(DecimalType.FIELD_TYPE, DecimalType.convert)
+
+        sqlite3.register_adapter(dict, DictType.adapt)
+        sqlite3.register_converter(DictType.FIELD_TYPE, DictType.convert)
 
         # turn on foreign keys
         # http://www.sqlite.org/foreignkeys.html
@@ -276,13 +275,6 @@ class SQLite(SQLInterface):
         if not self.connected: self.connect()
         return self._connection
 
-    def _get_thread(self):
-        if thread:
-            ret = str(thread.get_ident())
-        else:
-            ret = ""
-        return ret
-
     def _close(self):
         self._connection.close()
         self._connection = None
@@ -305,122 +297,7 @@ class SQLite(SQLInterface):
         ret = self._query(query_str, query_args, **kwargs)
         return [r['tbl_name'] for r in ret]
 
-    def get_field_SQL(self, field_name, field):
-        """
-        returns the SQL for a given field with full type information
-
-        http://www.sqlite.org/datatype3.html
-
-        field_name -- string -- the field's name
-        field -- Field() -- the set options for the field
-
-        return -- string -- the field type (eg, foo BOOL NOT NULL)
-        """
-        field_type = ""
-        is_pk = field.options.get('pk', False)
-        interface_type = field.interface_type
-
-        if issubclass(interface_type, bool):
-            field_type = 'BOOLEAN'
-
-        elif issubclass(interface_type, long):
-            if is_pk:
-                field_type = 'INTEGER PRIMARY KEY'
-            else:
-                field_type = 'BIGINT'
-
-        elif issubclass(interface_type, int):
-            field_type = 'INTEGER'
-            if is_pk:
-                field_type += ' PRIMARY KEY'
-
-        elif issubclass(interface_type, basestring):
-            fo = field.options
-            if field.is_ref():
-                # TODO -- 7-8-17 - this isn't a great way to do this, ideally the Field instance
-                # would combine all the options of both the current field and the
-                # foreign key field and return all those when Field.options is called
-                # (with the current field's options taking precedence) but there are
-                # lots of circular dependency things that happen when one field is
-                # trying to get the schema of another field and I don't have time
-                # to sort it all out right now
-                ref_s = field.schema
-                fo = ref_s.pk.options
-
-            if 'size' in fo:
-                field_type = 'CHARACTER({})'.format(fo['size'])
-            elif 'max_size' in fo:
-                field_type = 'VARCHAR({})'.format(fo['max_size'])
-            else:
-                field_type = 'TEXT'
-
-            if fo.get('ignore_case', False):
-                field_type += ' COLLATE NOCASE'
-
-            if is_pk:
-                field_type += ' PRIMARY KEY'
-
-        elif issubclass(interface_type, datetime.datetime):
-            #field_type = 'DATETIME'
-            field_type = 'TIMESTAMP'
-
-        elif issubclass(interface_type, datetime.date):
-            field_type = 'DATE'
-
-        elif issubclass(interface_type, float):
-            field_type = 'REAL'
-            size = field.options.get('size', field.options.get('max_size', 0))
-            if size > 6:
-                field_type = 'DOUBLE PRECISION'
-
-        elif issubclass(interface_type, decimal.Decimal):
-            field_type = 'NUMERIC'
-
-        elif issubclass(interface_type, bytearray):
-            field_type = 'BLOB'
-
-        else:
-            raise ValueError('Unknown python type: {}'.format(interface_type.__name__))
-
-        if field.required:
-            field_type += ' NOT NULL'
-        else:
-            field_type += ' NULL'
-
-        if not is_pk:
-            if field.is_ref():
-                ref_s = field.schema
-                if field.required: # strong ref, it deletes on fk row removal
-                    field_type += ' REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE CASCADE'.format(
-                        ref_s,
-                        ref_s.pk.name
-                    )
-
-                else: # weak ref, it sets column to null on fk row removal
-                    field_type += ' REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE SET NULL'.format(
-                        ref_s,
-                        ref_s.pk.name
-                    )
-
-        return '{} {}'.format(self._normalize_name(field_name), field_type)
-
-    def _set_table(self, schema, **kwargs):
-        """
-        http://sqlite.org/lang_createtable.html
-        """
-        query_str = []
-        query_str.append("CREATE TABLE {} (".format(self._normalize_table_name(schema)))
-
-        query_fields = []
-        for field_name, field in schema.fields.items():
-            query_fields.append('  {}'.format(self.get_field_SQL(field_name, field)))
-
-        query_str.append(",{}".format(os.linesep).join(query_fields))
-        query_str.append(')')
-        query_str = os.linesep.join(query_str)
-        ret = self._query(query_str, ignore_result=True, **kwargs)
-
-    def _set_index(self, schema, name, fields, **index_options):
+    def _set_index(self, schema, name, field_names, **index_options):
         """
         https://www.sqlite.org/lang_createindex.html
         """
@@ -429,7 +306,7 @@ class SQLite(SQLInterface):
             schema,
             name,
             self._normalize_table_name(schema),
-            ', '.join((self._normalize_name(f) for f in fields))
+            ', '.join(map(self._normalize_name, field_names))
         )
 
         return self._query(query_str, ignore_result=True, **index_options)
@@ -451,31 +328,6 @@ class SQLite(SQLInterface):
 
         return ret
 
-    def _insert(self, schema, fields, **kwargs):
-        """
-        http://www.sqlite.org/lang_insert.html
-        """
-        field_formats = []
-        field_names = []
-        query_vals = []
-        for field_name, field_val in fields.items():
-            field_names.append(self._normalize_name(field_name))
-            field_formats.append(self.val_placeholder)
-            query_vals.append(field_val)
-
-        query_str = "INSERT INTO {} ({}) VALUES ({})".format(
-            self._normalize_table_name(schema),
-            ', '.join(field_names),
-            ', '.join(field_formats)
-        )
-
-        ret = self._query(query_str, query_vals, cursor_result=True, **kwargs)
-
-        pk_name = schema.pk_name
-        # http://stackoverflow.com/questions/6242756/
-        # could also do _query('SELECT last_insert_rowid()')
-        return ret.lastrowid if pk_name not in fields else fields[pk_name]
-
     def _delete_tables(self, **kwargs):
         self._query('PRAGMA foreign_keys = OFF', ignore_result=True, **kwargs);
         ret = super(SQLite, self)._delete_tables(**kwargs)
@@ -487,31 +339,30 @@ class SQLite(SQLInterface):
         query_str = "DROP TABLE IF EXISTS {}".format(self._normalize_table_name(schema))
         ret = self._query(query_str, ignore_result=True, **kwargs)
 
-    def _handle_error(self, schema, e, **kwargs):
-        ret = False
+    def create_error(self, e, **kwargs):
         if isinstance(e, sqlite3.OperationalError):
             e_msg = str(e)
             if "no such column" in e_msg or "has no column" in e_msg:
                 #INSERT: "table yscrmiklbgdtx has no column named che"
                 #SELECT: "no such column: che"
-                try:
-                    ret = self._set_all_fields(schema, **kwargs)
-                except ValueError as e:
-                    ret = False
+                e = FieldError(e)
+
             elif "no such table" in e_msg:
-                ret = self._set_all_tables(schema, **kwargs)
+                e = TableError(e)
 
             elif "UNIQUE" in e_msg:
-                self.raise_error(e, e_class=UniqueError)
+                e = UniqueError(e)
 
-        return ret
+            else:
+                e = super().create_error(e, **kwargs)
+
+        elif isinstance(e, sqlite3.IntegrityError):
+            e = UniqueError(e)
 
-    def _create_error(self, e, exc_info):
-        if isinstance(e, sqlite3.IntegrityError):
-            er = UniqueError(e, exc_info)
         else:
-            er = super(SQLite, self)._create_error(e, exc_info)
-        return er
+            e = super().create_error(e, **kwargs)
+
+        return e
 
     def _get_fields(self, table_name, **kwargs):
         """return all the fields for the given table"""
@@ -540,11 +391,13 @@ class SQLite(SQLInterface):
             "BLOB": bytearray,
         }
 
+        for field_type, adapter in sqlite3.adapters.items():
+            if adapter_class := getattr(adapter, "__self__", None):
+                if adapter_class.FIELD_TYPE not in pg_types:
+                    pg_types[adapter_class.FIELD_TYPE] = field_type[0]
+
         # the rows we can set: field_type, name, field_required, min_size, max_size,
         #   size, unique, pk, <foreign key info>
-        # These keys will roughly correspond with schema.Field
-        # TODO -- we could actually use "type" to get the size because SQLite returns
-        # a value like VARCHAR[32]
         for row in fields:
             field = {
                 "name": row["name"],
@@ -557,10 +410,6 @@ class SQLite(SQLInterface):
                     field["field_type"] = ty
                     break
 
-            if field["pk"] and field["field_type"] is int:
-                # we compensate for SQLite internally setting pk to int
-                field["field_type"] = long
-
             if row["name"] in fks:
                 field["schema_table_name"] = fks[row["name"]]["table"]
                 field["ref_table_name"] = fks[row["name"]]["table"]
@@ -631,3 +480,39 @@ class SQLite(SQLInterface):
             offset
         )
 
+    def render_datatype_int_sql(self, field_name, field, **kwargs):
+        if field.is_pk():
+            field_type = 'INTEGER PRIMARY KEY'
+
+        else:
+            # we could break these up into tiny, small, and big but it
+            # doesn't really matter so we're not bothering
+            # https://www.sqlite.org/datatype3.html
+            size = field.size_info()["size"]
+
+            if size < 9223372036854775807:
+                field_type = 'INTEGER'
+
+            else:
+                field_type = NumericType.FIELD_TYPE
+
+        return field_type
+
+    def render_datatype_str_sql(self, field_name, field, **kwargs):
+            field_type = super().render_datatype_str_sql(field_name, field, **kwargs)
+
+            fo = field.interface_options
+            if fo.get('ignore_case', False):
+                field_type += ' COLLATE NOCASE'
+
+            return field_type
+
+    def render_datatype_datetime_sql(self, field_name, field, **kwargs):
+        return DatetimeType.FIELD_TYPE
+
+    def render_datatype_dict_sql(self, field_name, field, **kwargs):
+        return DictType.FIELD_TYPE
+
+    def render_datatype_uuid_sql(self, field_name, field, **kwargs):
+        return self.render_datatype_str_sql(field_name, field)
+
