@@ -4,13 +4,23 @@ import inspect
 import sys
 import datetime
 
-from datatypes.collections import Pool
+from datatypes import (
+    Pool,
+    EnglishWord,
+    NamingConvention,
+)
 
 # first party
 from .query import Query, Iterator
 from . import decorators, utils
 from .interface import get_interface
-from .config import Schema, Field, Index
+from .config import (
+    Schema,
+    Field,
+    AutoDatetime,
+    AutoIncrement,
+    Index
+)
 from .compat import *
 
 
@@ -35,7 +45,7 @@ class OrmPool(Pool):
         self.orm_class = orm_class
 
     def __missing__(self, pk):
-        o = self.orm_class.query.eq_field("pk", pk).one()
+        o = self.orm_class.query.eq_pk(pk).one()
         self[pk] = o
         return o
 
@@ -55,43 +65,6 @@ class Orm(object):
             che = Field(str, True)
 
             index_barche = Index('bar', 'che')
-
-        # create a more complex class using a field override
-        class Foo2(Orm):
-            table_name = "<TABLE NAME>"
-
-            class bar(Field):
-                type = int
-                required = True
-                options = {
-                    "default": 0,
-                    "unique": True,
-                    "max_size": 512,
-                }
-
-                def fget(self, orm, v):
-                    print("fget")
-                    return v
-
-                def iget(self, orm, v):
-                    print("iget")
-                    return v
-
-                def fset(self, orm, v):
-                    print("fset")
-                    return v
-
-                def fdel(self, orm, v):
-                    print("fdel")
-                    return v
-
-                def iquery(self, query, v):
-                    print("iquery")
-                    return v
-
-                def jsonable(self, orm, v):
-                    print("jsonable")
-                    return v
     """
     connection_name = ""
     """the name of the connection to use to retrieve the interface"""
@@ -102,43 +75,44 @@ class Orm(object):
     iterator_class = Iterator
     """the class this Orm will use for iterating through results returned from db"""
 
-    _id = Field(long, True, pk=True)
+    orm_classes = {}
+    """This will hold all other orm classes that have been loaded into memory
+    the class path is the key and the class object is the value"""
 
-    class _created(Field):
-        type = datetime.datetime
-        required = True
-        options = {
-            "aliases": ["created"],
-        }
-
-        def iset(self, orm, val):
-            if not val and orm.is_insert():
-                val = datetime.datetime.utcnow()
-            return val
-
-    class _updated(Field):
-        type = datetime.datetime
-        required = True
-        options = {
-            "aliases": ["updated"],
-        }
-
-        def iset(self, orm, val):
-            if val:
-                if not self.modified(orm, val):
-                    val = datetime.datetime.utcnow()
-
-            else:
-                val = datetime.datetime.utcnow()
-
-            return val
+    _id = AutoIncrement(aliases=["id"])
+    _created = AutoDatetime(created=True, updated=False, aliases=["created"])
+    _updated = AutoDatetime(created=False, updated=True, aliases=["updated"])
 
     @decorators.classproperty
     def table_name(cls):
-        return "{}_{}".format(
-            cls.__module__.lower().replace(".", "_"),
-            cls.__name__.lower()
-        )
+        return NamingConvention("{}_{}".format(
+            cls.__module__.replace(".", "_"),
+            cls.__name__
+        )).snakecase()
+
+    @decorators.classproperty
+    def model_name(cls):
+        """Returns the name for this orm
+
+        :Example:
+            class Collection(Orm): pass
+
+            print(Collection.model_name) # collection
+            print(Collection.models_name) # collections
+
+        :param plural: bool, True if you would like the plural model name
+        :returns: str, the model name in either singular or plural form
+        """
+        return NamingConvention(cls.__name__).snakecase()
+
+    @decorators.classproperty
+    def models_name(cls):
+        """Returns the plural name for this orm
+
+        :see: .model_name
+        :returns: str, the model name in plural form
+        """
+        return EnglishWord(cls.model_name).plural()
 
     @decorators.classproperty
     def schema(cls):
@@ -167,6 +141,11 @@ class Orm(object):
         query_class = cls.query_class
         return query_class(orm_class=cls)
 
+    @property
+    def field_names(self):
+        """Return all the field names"""
+        return [k for k in self.schema.fields]
+
     @property
     def fields(self):
         """
@@ -178,12 +157,18 @@ class Orm(object):
         return {k:getattr(self, k, None) for k in self.schema.fields}
 
     @property
-    def modified_fields(self):
-        modified_fields = set()
+    def modified_field_names(self):
+        """Return all the field names that are currently modified"""
+        modified_field_names = set()
         for field_name, field in self.schema.fields.items():
             if field.modified(self, getattr(self, field_name)):
-                modified_fields.add(field_name)
-        return modified_fields
+                modified_field_names.add(field_name)
+        return modified_field_names
+
+    @property
+    def modified_fields(self):
+        """Return a dict of field_names/field_values for all the currently modified fields"""
+        return {k:getattr(self, k) for k in self.modified_field_names}
 
     @classmethod
     def pool(cls, maxsize=0):
@@ -225,7 +210,7 @@ class Orm(object):
         return instance
 
     @classmethod
-    def make_dict(cls, fields, fields_kwargs):
+    def make_dict(cls, fields, fields_kwargs, schema=None):
         """Lots of methods take a dict and key=val for fields, this combines fields
         and fields_kwargs into one master dict, turns out we want to do this more
         than I would've thought to keep api compatibility with prom proper
@@ -233,9 +218,18 @@ class Orm(object):
         :param fields: dict, the fields in a dict
         :param fields_kwargs: dict, if you would like to pass the fields as key=val
             this picks those up and combines them with fields
+        :schema: Schema, if passed in then this will normalize field names
         :returns: dict, the combined fields
         """
-        return utils.make_dict(fields, fields_kwargs)
+        fields = utils.make_dict(fields, fields_kwargs)
+
+        if schema:
+            # since schema is passed in resolve any aliases
+            for field_name in list(fields.keys()):
+                if fn := schema.field_name(field_name, ""):
+                    fields[fn] = fields.pop(field_name)
+
+        return fields
 
     def __init__(self, fields=None, **fields_kwargs):
         """Create an Orm object
@@ -246,20 +240,24 @@ class Orm(object):
         self._interface_pk = None
         self._interface_hydrate = False
 
-        fields = self.make_dict(fields, fields_kwargs)
-
-        # resolve any aliases
-        for fn in list(fields.keys()):
-            if self.schema.has_field(fn):
-                field_name = self.schema.field_name(fn)
-                fields[field_name] = fields.pop(fn)
+        schema = self.schema
+        fields = self.make_dict(fields, fields_kwargs, schema=schema)
 
         # set defaults
-        for field_name, field in self.schema.fields.items():
+        for field_name, field in schema.fields.items():
             fields[field_name] = field.fdefault(self, fields.get(field_name, None))
 
         self.modify(fields)
 
+    def __init_subclass__(cls):
+        """When this class is loaded into memory it will be saved into cls.orm_classes,
+        this way every orm class knows about all the others
+
+        https://peps.python.org/pep-0487/
+        """
+        super().__init_subclass__()
+        cls.orm_classes[f"{cls.__module__}:{cls.__name__}"] = cls
+
     def fk(self, orm_class):
         """find the field value in self that is the primary key of the passed in orm_class
 
@@ -301,8 +299,7 @@ class Orm(object):
 
     def is_update(self):
         """Return True if .save() will perform an interface update"""
-        pk = self._interface_pk
-        return pk is not None
+        return self._interface_pk is not None
 
     def is_insert(self):
         """Return True if .save() will perform an interface insert"""
@@ -351,7 +348,7 @@ class Orm(object):
 
         return fields
 
-    def insert(self):
+    def insert(self, **kwargs):
         """persist the field values of this orm"""
         ret = True
 
@@ -360,7 +357,7 @@ class Orm(object):
 
         q = self.query
         q.set(fields)
-        pk = q.insert()
+        pk = q.insert(**kwargs)
         if pk:
             fields = q.fields_set.fields
             pk_name = schema.pk_name
@@ -373,7 +370,7 @@ class Orm(object):
 
         return ret
 
-    def update(self):
+    def update(self, **kwargs):
         """re-persist the updated field values of this orm that has a primary key"""
         ret = True
         fields = self.to_interface()
@@ -383,12 +380,12 @@ class Orm(object):
 
         pk = self._interface_pk
         if pk:
-            q.is_field(self.schema.pk.name, pk)
+            q.eq_field(self.schema.pk.name, pk)
 
         else:
             raise ValueError("Cannot update an unhydrated orm instance")
 
-        if q.update():
+        if q.update(**kwargs):
             fields = q.fields_set.fields
             self.from_interface(fields)
 
@@ -397,7 +394,46 @@ class Orm(object):
 
         return ret
 
-    def save(self):
+    def upsert(self, **kwargs):
+        """Perform an UPSERT query where we insert the fields if they don't already
+        exist on the db or we UPDATE if they do
+
+        We only want to upsert on specific occasions where we know we've set the
+        conflict values and will be sending them to the db. UPSERT queries need
+        to have a unique index on the table they can use for the conflict fields.
+
+        This method will go through the indexes and try and find a unique index
+        that has all fields that are being sent to the interface and it will use
+        those fields as the conflict fields, it will raise a ValueError if it can't
+        find a valid set of conflict fields
+
+        :param **kwargs: passed through to the interface
+        """
+        ret = True
+
+        schema = self.schema
+        fields = self.to_interface()
+
+        conflict_fields = self.conflict_fields(fields)
+        if not conflict_fields:
+            raise ValueError(f"Upsert failed to find the conflict field names")
+
+        q = self.query
+        q.set(fields)
+        pk = q.upsert([t[0] for t in conflict_fields], **kwargs)
+        if pk:
+            fields = q.fields_set.fields
+            pk_name = schema.pk_name
+            if pk_name:
+                fields[pk_name] = pk
+                self.from_interface(fields)
+
+        else:
+            ret = False
+
+        return ret
+
+    def save(self, **kwargs):
         """
         persist the fields in this object into the db, this will update if _id is set, otherwise
         it will insert
@@ -408,20 +444,20 @@ class Orm(object):
 
         pk = self._interface_pk
         if pk:
-            ret = self.update()
+            ret = self.update(**kwargs)
         else:
-            ret = self.insert()
+            ret = self.insert(**kwargs)
 
         return ret
 
-    def delete(self):
+    def delete(self, **kwargs):
         """delete the object from the db if pk is set"""
         ret = False
         q = self.query
         pk = self._interface_pk
         if pk:
             pk_name = self.schema.pk.name
-            self.query.is_field(pk_name, pk).delete()
+            self.query.eq_field(pk_name, pk).delete(**kwargs)
 
             for field_name, field in self.schema.fields.items():
                 setattr(self, field_name, field.idel(self, getattr(self, field_name)))
@@ -433,13 +469,89 @@ class Orm(object):
 
         return ret
 
+    def conflict_fields(self, fields):
+        """Internal method. This will find fields that can be used for .upsert/.load
+
+        :param fields: dict, the fields to check for values that would satisfy
+            unique indexes or a primary key
+        :returns: list<tuple>, a list of (field_name, field_value) tuples
+        """
+        schema = self.schema
+
+        conflict_fields = []
+
+        # we'll start with checking the primary key
+        for field_name in schema.pk_names:
+            if field_name in fields:
+                conflict_fields.append((field_name, fields[field_name]))
+
+            else:
+                conflict_fields = []
+                break
+
+        if not conflict_fields:
+            # no luck with the primary key, so let's check unique indexes 
+            for index in schema.indexes.values():
+                if index.unique:
+                    for field_name in index.field_names:
+                        if field_name in fields:
+                            conflict_fields.append((field_name, fields[field_name]))
+
+                        else:
+                            conflict_fields = []
+                            break
+
+        return conflict_fields
+
+    def load(self):
+        """Given a partially populated orm try and load any missing fields from
+        the db
+
+        :returns: bool, True if it loaded from the db, False otherwise
+        """
+        fields = self.modified_fields
+        conflict_fields = self.conflict_fields(fields)
+        if not conflict_fields:
+            raise ValueError(f"Load failed to find suitable fields to query on")
+
+        q = self.query
+        for field_name, field_val in conflict_fields:
+            q.eq_field(field_name, field_val)
+
+        field_names = []
+        for field_name in self.schema.fields.keys():
+            if field_name not in fields:
+                field_names.append(field_name)
+
+        q.select(*field_names)
+        field_values = q.one()
+        if field_values:
+            ret = True
+            fields = dict(zip(field_names, field_values))
+            self.from_interface(fields)
+
+            # can't decide if I should actually set this or not
+            self._interface_hydrate = True
+
+        else:
+            ret = False
+
+        return ret
+
     def requery(self):
         """Fetch this orm from the db again (ie, re-query the row from the db and
         return a new Orm instance with the columns from that row)"""
-        pk = self._interface_pk
-        if not pk:
-            raise ValueError("Unable to refetch orm via hydrated primary key")
-        return self.query.eq_pk(pk).one()
+        fields = {k:v for k, v in self.fields.items() if v is not None}
+
+        conflict_fields = self.conflict_fields(fields)
+        if not conflict_fields:
+            raise ValueError("Unable to refetch orm")
+
+        q = self.query
+        for field_name, field_val in conflict_fields:
+            q.eq_field(field_name, field_val)
+
+        return q.one()
 
     def is_modified(self, field_name=""):
         """true if a field, or any field, has been changed from its original value, false otherwise
@@ -448,9 +560,9 @@ class Orm(object):
         :returns: bool
         """
         if field_name:
-            ret = field_name in self.modified_fields
+            ret = field_name in self.modified_field_names
         else:
-            ret = len(self.modified_fields) > 0
+            ret = len(self.modified_field_names) > 0
         return ret
 
     def modify(self, fields=None, **fields_kwargs):
@@ -465,10 +577,11 @@ class Orm(object):
         """
         fields = self.make_dict(fields, fields_kwargs)
         fields = self.modify_fields(fields)
+        schema = self.schema
 
         for field_name, field_val in fields.items():
-            if self.schema.has_field(field_name):
-                setattr(self, field_name, field_val)
+            if fn := schema.field_name(field_name, None):
+                setattr(self, fn, field_val)
 
     def modify_fields(self, fields):
         """In child classes you should override this method to do any default 
@@ -486,15 +599,21 @@ class Orm(object):
             field_name = self.schema.field_name(k)
 
         except AttributeError:
-            # we treat pk (alias for _id) as special because we usually want the
-            # pk to just return None, even if it doesn't exist, the reason why is
-            # because usually you remove the pk by setting `OrmClass._id = None`
-            # and so self._id would return None, so we want self.pk to return
-            # None also. This should only ever be checked if the field doesn't
-            # exist, and self.pk won't exist if there is no primary key
-            # https://github.com/Jaymon/prom/issues/139#issuecomment-944806055
-            if k != "pk":
-                raise
+            for ref_field_name, ref_field in self.schema.ref_fields.items():
+                ref_class = ref_field.ref
+                if k == ref_class.model_name:
+                    ref_field_value = getattr(self, ref_field_name, None)
+                    ret = None
+                    if ref_field_value:
+                        ret = ref_class.query.eq_pk(ref_field_value).one()
+
+                    return ret
+
+            raise
+
+            # ??? - could we do something with models_name here? Go through all
+            # the orm_classes looking for a models_name match and querying them
+            # using self.pk?
 
         else:
             ret = getattr(self, field_name)
@@ -504,9 +623,17 @@ class Orm(object):
     def __setattr__(self, k, v):
         try:
             field_name = self.schema.field_name(k)
+
         except AttributeError:
+            for ref_field_name, ref_field in self.schema.ref_fields.items():
+                ref_class = ref_field.ref
+                if k == ref_class.model_name:
+                    k = ref_field_name
+                    v = v.pk
+                    break
+
             field_name = k
-        return super(Orm, self).__setattr__(field_name, v)
+        return super().__setattr__(field_name, v)
 
     def __delattr__(self, k):
         try:
