@@ -66,7 +66,17 @@ class Iterator(ListIterator):
         ret = False
         if self.query.bounds.has_more():
             cursor = self.cursor()
-            ret = self.query.bounds.limit_paginate == cursor.rowcount
+            if cursor.rowcount == -1:
+                try:
+                    if self[self.query.bounds.find_more_index()]:
+                        ret = True
+
+                except IndexError:
+                    pass
+
+            else:
+                ret = self.query.bounds.limit_paginate == cursor.rowcount
+
         return ret
 
     def cursor(self):
@@ -90,11 +100,7 @@ class Iterator(ListIterator):
 
     def next(self):
         cursor = self.cursor()
-
-        if is_py2:
-            cursor_next = cursor.next
-        else:
-            cursor_next = cursor.__next__
+        cursor_next = cursor.__next__
 
         # if we have paginated the results we have to account for requesting the
         # one extra row to see if we have more results waiting
@@ -223,12 +229,18 @@ class Iterator(ListIterator):
             one field then just that value will be returned
         """
         r = None
+        orm_class = self.orm_class
         if self.field_names:
-            field_vals = [d.get(fn, None) for fn in self.field_names]
+            field_vals = []
+            for field_name in self.field_names:
+                fv = d.get(field_name, None)
+                if orm_class:
+                    fv = orm_class.schema.fields[field_name].iget(None, fv)
+                field_vals.append(fv)
+            #field_vals = [d.get(fn, None) for fn in self.field_names]
             r = field_vals if len(self.field_names) > 1 else field_vals[0]
 
         else:
-            orm_class = self.orm_class
             if orm_class:
                 r = orm_class.hydrate(d)
             else:
@@ -329,7 +341,6 @@ class Bounds(object):
 
     def __bool__(self):
         return self.limit > 0 or self.offset > 0
-    __nonzero__ = __bool__ # py2
 
     def has(self):
         return bool(self)
@@ -377,6 +388,10 @@ class Bounds(object):
 
         return offset
 
+    def find_more_index(self):
+        #return self.offset + self.limit_paginate
+        return self.offset + self.limit
+
 
 class Field(object):
     @property
@@ -725,12 +740,12 @@ class Query(object):
 
     def is_field(self, field_name, field_val=None, **field_kwargs):
         return self.append_operation("eq", field_name, field_val, **field_kwargs)
-    def eq_field(self, field_name, field_val, **field_kwargs):
+    def eq_field(self, field_name, field_val=None, **field_kwargs):
         return self.is_field(field_name, field_val, **field_kwargs)
 
     def not_field(self, field_name, field_val=None, **field_kwargs):
         return self.append_operation("ne", field_name, field_val, **field_kwargs)
-    def ne_field(self, field_name, field_val, **field_kwargs):
+    def ne_field(self, field_name, field_val=None, **field_kwargs):
         return self.not_field(field_name, field_val, **field_kwargs)
 
     def between_field(self, field_name, low, high):
@@ -823,6 +838,9 @@ class Query(object):
         self._ifilter = predicate
         return self
 
+    def filter(self, predicate):
+        return self.ifilter(predicate)
+
     def limit(self, limit):
         self.bounds.limit = limit
         return self
@@ -914,21 +932,41 @@ class Query(object):
     def exists(self):
         return self.has()
 
-    def insert(self):
+    def insert(self, **kwargs):
         """persist the .fields"""
-        return self.interface.insert(self.schema, self.fields_set.fields)
+        return self.interface.insert(self.schema, self.fields_set.fields, **kwargs)
 
-    def update(self):
+    def update(self, **kwargs):
         """persist the .fields using .fields_where"""
         return self.interface.update(
             self.schema,
             self.fields_set.fields,
-            self
+            self,
+            **kwargs
         )
 
-    def delete(self):
+    def upsert(self, conflict_field_names=None, **kwargs):
+        """persist the .fields"""
+        insert_fields = self.fields_set.fields
+        update_fields = dict(insert_fields)
+
+        if not conflict_field_names:
+            conflict_field_names = self.schema.pk_names
+
+        for field_name in conflict_field_names:
+            update_fields.pop(field_name, None)
+
+        return self.interface.upsert(
+            self.schema,
+            insert_fields,
+            update_fields,
+            conflict_field_names,
+            **kwargs
+        )
+
+    def delete(self, **kwargs):
         """remove fields matching the where criteria"""
-        return self.execute('delete')
+        return self.execute('delete', **kwargs)
 
     def render(self, **kwargs):
         """Render the query
@@ -972,10 +1010,6 @@ class Query(object):
                 setattr(instance, key, copy.deepcopy(val, memodict))
         return instance
 
-    def __unicode__(self):
+    def __str__(self):
         return self.render()
 
-    def __str__(self):
-        ret = self.__unicode__()
-        return ByteString(ret) if is_py2 else ret
-
